#
# (C) Copyright 2000,2001 Timothy Robert Ansell
# All Rights Reserved
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# Author: Tim Ansell (ak mithro) mithro@mithis.net

#
# This class only authenticates users, it stores no properties.
#

from Globals import DTMLFile, Persistent, HTMLFile
from OFS.Folder import Folder
from BTree import BTree
import Acquisition

import string
import socket
import sys
import time

import crypt

DEBUG = 1

if sys.version[:5] == '1.5.2' :
	import re
else:
	import re


manage_cacheForm=HTMLFile('manage_cacheForm', globals())

class ExternalBase(Folder, Persistent):
	""" This is the basic class for most socketed based authentication  """

	#####################################################################################################
	#####################################################################################################
	#####################################################################################################
	
	### These stuff define what this connection waits for. They are all regexs ###
	# [Subgroup that has the bit, leftovers, regex]
	connectstart = None
	sendusername = None
	sendpassword = None
	passwordok = None
	passwordnotokay = None
	okaytosendagain = None
	error = None
	
	# This stuff is what is sent to the user
	username = None
	password = None
	after = None

	# Functions that can be overriden
	def postInitialisation(self, REQUEST):
		pass

	def finishUp(self, data, socket, username):
		"""
		This function takes remaining data and the socket and username.
		It's called after a successful user authenticate
		"""
		return

	def vertifyUser(self, username):
		""" This function takes a username and says if it still exists"""
		return 1

	def manage_editAuthSourceAfter(self, REQUEST):
		pass
		
	#####################################################################################################
	#####################################################################################################
	#####################################################################################################	

	manage_cache = manage_cacheForm
	def get_keys_cache(self):
		return self._cache.keys()
	def get_keys_bad(self):
		return self._bad_cache.keys()

	meta_type='Authorisation Source'
	title='External Base Authentication'

	icon ='misc_/exUserFolder/exUserFolderPlugin.gif'

	manage_tabs=Acquisition.Acquired

	#
	# You can define this to go off and do the authentication instead of
	# using the basic one inside the User Object
	#
	def remoteAuthMethod(self, username, password):
		""" This method connects to a server and trys to authenticate against it """
		self._p_changed = 1

		gP = self.currentPropSource.getUserProperty
		sP = self.currentPropSource.setUserProperty

		#if self.cache_time + float(gP(username=username, key='_expire', default=0)) > time.time():
		#
		#	if crypt.crypt(password, username) == gP(username=username, key='_goodpassword', default=None):
		#		debug( "Cache: Okay" )
		#		return 1

		bad = self._bad_cache.get(username, None)
		if bad:
 			if bad[0] == crypt.crypt(password, username) and bad[1] + self.cache_time > time.time():
				debug( "Cache: Not Okay" )
				return 0

		# Gets the socket
		thesocket = self.checkOpen()

		data = thesocket.recv(1)

		connectionstarted = 0
		sendusername = 0
		sendpassword = 0

		start_time = time.time()

		while data and time.time() < start_time + self.time_out:

			debug( "Data = %s" % (data) )
			
			if connectionstarted:

				if not sendusername:
					# Check if it's okay to send the username
					matches = self.sendusername[2].match(data)
					if matches:
						data = matches.group(self.sendusername[1])
						sendusername = 1

						debug("sending: %s" % (self.username % (username)) )
						thesocket.send(self.username % (username))

				elif not sendpassword:
					# Check if it's okay to send the password
					matches = self.sendpassword[2].match(data)
					if matches:
						data = matches.group(self.sendpassword[1])
						sendpassword = 1

						debug("sending: %s" % (self.password % (password)) )
						thesocket.send(self.password % (password))
				else:
					# Wait for okay or not okay
					matches = self.passwordok[2].match(data)
					
					if matches:
						data = matches.group(self.passwordok[1])

						thesocket.send(self.after)

						if self.store_users:
							self._cache[username] = 1

						# Cache the stuff
						sP(username=username, key='_expire', value=str(time.time()))
						sP(username=username, key='_goodpassword', value=crypt.crypt(password, username))

						
						print "Calling finishUp(%s, %s, %s)" % (data, thesocket, username)
						self.finishUp(data, thesocket, username)
						try:
							pass
						except x:
							debug("Error in finishup: %s" % (x))

						debug( "Not-cache: Good password!" )

						# Update ourselves
						self._p_changed = 1
						get_transaction().commit()
						
						return 1

					matches = self.passwordnotokay[2].match(data)
					
					if matches:
						data = matches.group(self.passwordnotokay[1])

						thesocket.send(self.after)

						if self.store_users:
							self._bad_cache[username] = [time.time(), crypt.crypt(password,username)]

							# Update ourselves
							self._p_changed = 1
							get_transaction().commit()

						debug( "Not-cache: Bad password!" )
						return 0

			else:
				# Check if the connection has started
				matches = self.connectstart[2].match(data)
				if matches:
					data = matches.group(self.connectstart[1])
					connectionstarted = 1

			# Check for an error
			matches = self.error[2].match(data)
			if matches:
				debug( "Error!" )
				return 0

			data = data + thesocket.recv(1)
			
		debug( "Timeout!" )
		return 0
	
	def __init__(self, id, host, port, cache_time=6000, time_out=600, store_users=1):
		self.id = id
		self.host = str(host)
		self.port = int(port)
		self.cache_time = int(cache_time)
		self.time_out = int(time_out)
		self.store_users = store_users

		# These store user information
		if self.store_users:
			self._cache = BTree()
			self._bad_cache = BTree()
		
		self._p_changed = 1

	def manage_editAuthSource(self, REQUEST):
		""" Edit the authentication stats """
		self.host = str(REQUEST['host'])
		self.port = int(REQUEST['port'])
		self.cache_time = int(REQUEST['cache_time'])
		self.time_out = int(REQUEST['time_out'])

		self.manage_editAuthSourceAfter(REQUEST)

		debug("Just set the attributes")
		
		self._p_changed = 1

	def checkOpen(self):
		""" Open a socket """
		rsocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		debug( "connecting to %s %s" % (self.host, self.port) )
		rsocket.connect( (self.host, self.port) )
		
		return rsocket

	def createUser(self, username, password, roles):
		""" Add A Username """
		pass

	def updateUser(self, username, password, roles):
		rolestring=''
		try:
			for role in roles:
				rolestring=rolestring+role+','
		except:
			pass

		rolestring=rolestring[:-1]
		self.currentPropSource.setUserProperty(username=username, key='_roles',
						       value=rolestring)
		
	def deleteUsers(self, userids):
		for user in userids:
			del self._cache[user]

	def listUsers(self):
		"""Returns a list of user names or [] if no users exist"""
		if self.store_users:
			list = []
			for i in self._cache.keys():
				if self.vertifyUser(i):
					# Build a list to be returned
					list.append({'username':i})
				else:
					# Clean up the users...
					self.currentAuthSource.deleteUsers([i])
					self.currentPropSource.deleteUsers([i])
			return list
		else:
			return []

	def listOneUser(self,username):
		users = []
		
		roles=[]			
		#roleString=self.currentPropSource.getUserProperty(username=username, key='_roles')
		#if roleString:
		#	roles=string.split(roleString,',')
		#	
		N={'username':username, 'password':'', 'roles':roles}
		users.append(N)
			
		return users
	
	def getUsers(self):
		"""Return a list of user objects or [] if no users exist"""
		data=[]
		try:    items=self.listusers()
		except: return data



def debug(string):
	if DEBUG:
		print string
	
	
